#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>
#include <sys/stat.h>
#include <sys/sem.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include "common.h"

#define MESSAGE_QUEUE_KEY 1347

typedef struct msgBuf
{
        long msgType;
        char name[11];
        char text[256];
}msgBuf;

// 공유 메모리  연결
void ConnectSharedMemory();

// 공유메모리에 현재 프로세스 정보 저장
void WriteCurrentProcess(void **pShemBuffer, char *name);

// 메시지 큐를 생성 혹은 연결
void CreateOrConnectMessageQueue();

// 공유메모리에서 현재 프로세스 분리
void DetachSharedMemory();

// 메시지 큐 종료
void CloseMessageQueue();

// 메시지 전달 함수
void SendMessage(char *name);

// 세마포어 연결
void ConnectSemaphore();

// 세마포어 언락
void Semaphore_Unlock();

// 세마포어 락
void Semaphore_Lock();

// 시그널 등록
void InitializeSignal();

// 종료 시그널 핸들러 Ctrl + C, Stop Command 
void SIGINT_Handler(int signo);

// 유저 2명이 채워질 경우 상대방 접속 표시 및 채팅 가능 핸들러 
void SIGUSR1_Handler(int signo);

// 상대방에게 채팅이왔다는 핸들러
void SIGUSR2_Handler(int signo);

// 채팅 불가 핸들러
void SIGALRM_Handler(int signo);

int shared_memory_id = 0;
int msg_queue_id = 0;
int semaphore_id = 0;
void *pShemBuffer = NULL;
int isChattingPossible = FALSE;

int main(int argc, char *argv[])
{
	if(argc !=  2)
	{
		printf("you must input one argument\n");
		exit(1);
	}

	if(strlen(argv[1]) > 10)
	{
		printf("이름은 10글자를 넘어서는 안됩니다\n");
		exit(1);
	}

	InitializeSignal();

	ConnectSharedMemory();	

	ConnectSemaphore();

	WriteCurrentProcess(&pShemBuffer, argv[1]);

	CreateOrConnectMessageQueue();

	printf("[%s] client connect sucess\n", argv[1]);

	while(TRUE)
	{
		SendMessage(argv[1]);
	}
	
	return 0;
}

void CreateOrConnectMessageQueue()
{
	// 첫번째 클라이언트가 메시지 큐 생성
	msg_queue_id = msgget(MESSAGE_QUEUE_KEY, IPC_EXCL | IPC_CREAT | 0666);

	// 메시지큐가 이미  존재한다면
	if(-1 == msg_queue_id)
	{	
		// 기존 메시지 큐에 접속
		msg_queue_id = msgget(MESSAGE_QUEUE_KEY, IPC_CREAT | 0666);

		if(-1 == msg_queue_id)
		{
			printf("msg queue Connect Fail\n");

			DetachSharedMemory();
			
			exit(1);
		}
	}
}

void ConnectSharedMemory()
{
	shared_memory_id = shmget(SHARED_MEMORY_KEY, SHARED_MEMORY_SIZE, 0644 | IPC_EXCL | IPC_CREAT);
	
	// Shared Memory가 이미 생성되었는지 확인
	if(-1 == shared_memory_id)
        {
		// Shared Memory 연결
               	shared_memory_id = shmget(SHARED_MEMORY_KEY, SHARED_MEMORY_SIZE, 0644 | IPC_CREAT);

                if(-1 == shared_memory_id)
                {
                        printf("Shared Memory Connect Fail\n");
                        exit(1);
                }
        }
        else
        {
		if(-1 == shmctl(shared_memory_id , IPC_RMID, 0))
		{
			printf("ConnectSharedMemory shmctl error\n");	
		}

                printf("server doesn't exist\n");
                exit(1);
        }

	
        pShemBuffer = shmat(shared_memory_id, (void*)0, 0);
 
        if((void*)(-1) == pShemBuffer)
        {
                printf("shmat error\n");
                exit(1);
        }	
}

// 현재 프로세스 데이터를 공유메모리에 저장
void WriteCurrentProcess(void **pShemBuffer, char *processName)
{
        int i;
	SharedData *pSharedData = ((SharedData*)*pShemBuffer);

	if(NULL == pSharedData)
	{
		printf("SharedData Pointer is NULL\n");

		DetachSharedMemory();
		CloseMessageQueue();
	
		exit(1);
	}
	
	// 현재 접속 수를 조사하여 예외처리	
        if(pSharedData->numOfProcess == 2)
        {
                printf("process is full in server\n");

		DetachSharedMemory();

                exit(1);
        }

	// 현재 존재하고 있는 프로세스중에서 같은 이름이 존재하는지 조사
        for(i = 0;i < pSharedData->numOfProcess; i++)
        {
                if(strcmp(pSharedData->pData[i].name, processName) == 0)
                {
                        printf("%s name already exists\n", processName);
                        exit(1);
                }
        }
	
	Semaphore_Lock();

	// 접속 순서대로 차례대로 프로세스 데이터를 공유메모리에 저장
        pSharedData->pData[pSharedData->numOfProcess].pid = getpid();
        strcpy(pSharedData->pData[pSharedData->numOfProcess].name, processName);
        pSharedData->numOfProcess++;

	Semaphore_Unlock();	

	// 서버 측으로 접속 시그널 전달
	kill(pSharedData->Server.pid, SIGUSR1);
}

// 현재 프로세스를 공유메모리부터 분리
void DetachSharedMemory()
{
        if(shmdt(pShemBuffer) == -1)
        {
              	printf("shmdt error\n");
                exit(1);
        }
}

// 상대방에게 메시지 전달
void SendMessage(char *name)
{
	SharedData *pSharedData = (SharedData*)pShemBuffer;
	int length = 0;
	int i = 0;
	msgBuf buf;

	if(NULL == pSharedData)
	{
		printf("SharedData is NULL\n");
		return;
	}

	memset(&buf, 0, sizeof(buf));

        strcpy(buf.name, name);
        buf.msgType = getpid();

        while(isChattingPossible && NULL != fgets(buf.text, sizeof(buf.text), stdin))
        {
                length = strlen(buf.text);

                if(buf.text[length - 1] == '\n')
                {
                        buf.text[length - 1] = '\0';
                }

                if(msgsnd(msg_queue_id, &buf, length + 12, 0) == -1)
                {
                        printf("msgsnd error\n");
                }
                else
                {
			// 메시지 전달 성공 시에 상대방에게 메시지 전달했다는 신호 전달
         		for(i = 0; i < pSharedData->numOfProcess; i++)
			{
				if(pSharedData->pData[i].pid != getpid())
				{
					kill(pSharedData->pData[i].pid, SIGUSR2);	
					break;
				}
			}		               
                }
        }
}

// 클라이언트 종료 핸들러
void SIGINT_Handler(int signo)
{
	int i = 0;
	int index = 0;
	int pid = getpid();
	int serverPid = 0;
	SharedData *pSharedData = (SharedData*)pShemBuffer;

	if(NULL == pSharedData)
	{
		printf("SharedData is NULL\n");

		DetachSharedMemory();
	
		exit(1);
	}

	Semaphore_Lock();

	serverPid = pSharedData->Server.pid;
 
	// 종료하는 클라이언트가 SharedMemory 프로세스 데이터배열인덱스 0이라면
	// 1번 인덱스와 교체함으로써 종료하는 프로세스의 인덱스가 numOfProcess와 일치하도록 구성
	if(pSharedData->numOfProcess == 2 && pSharedData->pData[0].pid == getpid())
	{
		ProcessData tmp = pSharedData->pData[0];
		pSharedData->pData[0] = pSharedData->pData[1];
		pSharedData->pData[1]= tmp;
	}

	pSharedData->numOfProcess--; 

	Semaphore_Unlock();

	// 마지막의 클라이언트 종료하는 프로세스가 종료
	if(pSharedData->numOfProcess == 0)
	{
		CloseMessageQueue();	
	}

	DetachSharedMemory();

	// 서버에게 종료소식 알림
	kill(serverPid, SIGUSR2);

	// SendMessage에서 fgets으로 입력 대기중이므로 강제종료
	exit(1);
}

void CloseMessageQueue()
{
	if(msgctl(msg_queue_id, IPC_RMID, NULL) == -1)
	{
		fprintf(stderr, "msgctl error");
		exit(1);
	}
}

// 유저 2명이 채워질 경우 상대방 접속 표시 및 채팅 가능 핸들러
void SIGUSR1_Handler(int signo)
{
	SharedData *pSharedData = (SharedData*)pShemBuffer;
	int i = 0;

	if(NULL == pSharedData)
	{
		fprintf(stderr, "SharedData is NULL\n");
		exit(1);
	}
	
	for(i = 0; i < pSharedData->numOfProcess; i++)
	{
		if(pSharedData->pData[i].pid != getpid())
		{
			printf("[%s]가 접속했습니다.\n", pSharedData->pData[i].name);
			break;
		}
	}

	printf("채팅이 가능합니다.\n");

	isChattingPossible = TRUE;
}

// 상대방에게서 채팅이 왔다는 핸들러
void SIGUSR2_Handler(int signo)
{
	SharedData *pSharedData = (SharedData*)pShemBuffer;

	if(NULL == pSharedData)
	{
		fprintf(stderr, "pSharedData is NULL\n");
		exit(1);
	}

	msgBuf buf;
	int rcvMsgType = 0;
	int i;

	// 상대방의 pid를 받는 메시지 타입으로 지정
	for(i = 0; i < pSharedData->numOfProcess; i++)
	{
		if(pSharedData->pData[i].pid != getpid())
		{
			rcvMsgType = pSharedData->pData[i].pid;
			break;
		}
	}

	memset(&buf, 0, sizeof(buf));

	// 메시지 받은 후 출력
	if(-1 !=  msgrcv(msg_queue_id, &buf,sizeof(buf) - sizeof(long), rcvMsgType, 0))
        {
		printf("[%s] %s\n", buf.name, buf.text);
	}
}

// 상대방이 나갔을 시에 처리할 핸들러
void SIGALRM_Handler(int signo)
{
	SharedData *pSharedData = (SharedData*)pShemBuffer;
	int i = 0;

	isChattingPossible = FALSE;

	if(NULL == pSharedData)
	{
		printf("SharedData is NULL\n");
		return;
	}

	// 상대방의 정보를 출력
	for(i = 0; i < 2; i++)
	{
		if(pSharedData->pData[i].pid != getpid())
		{
			printf("[%s]가 채팅방에서 나갔습니다.\n", pSharedData->pData[i].name);
			break;
		}
	}
}

void InitializeSignal()
{
	if(SIG_ERR == signal(SIGTERM, SIGINT_Handler))
	{
		printf("cannot handle SIG_TERM\n");
		exit(EXIT_FAILURE);
	}
		
	if(SIG_ERR == signal(SIGINT, SIGINT_Handler)) 
	{
		printf("cannot handle SIGINT\n");
		exit(EXIT_FAILURE);
	}

	if(SIG_ERR == signal(SIGUSR1, SIGUSR1_Handler))
	{
		printf("cannot handle SIGUSR1\n");
		exit(EXIT_FAILURE);
	}
	if(SIG_ERR == signal(SIGUSR2, SIGUSR2_Handler))
	{
		printf("cannot handler SIGUSR2\n");
		exit(EXIT_FAILURE);
	}

	if(SIG_ERR == signal(SIGALRM, SIGALRM_Handler))
	{
		printf("cannot handler SIGALRM\n");
		exit(EXIT_FAILURE);
	}
}


void ConnectSemaphore()
{
	semaphore_id = semget(SEMAPHORE_KEY, 1, IPC_CREAT | IPC_EXCL | 0666);
	
	if(-1 == semaphore_id)
	{
		semaphore_id = semget(SEMAPHORE_KEY, 1, IPC_CREAT);

		if(-1 == semaphore_id)
		{
			printf("can't connect smeaphore\n");
		}
	}
	else
	{
		if(-1 == shmctl(semaphore_id, 0, IPC_RMID))
		{
			printf("semaphore delete fail\n");
		}

		printf("semaphore connect fail\n");

		exit(1);
	}
}

void Semaphore_Unlock()
{
	struct sembuf operation;

	operation.sem_num = 0;
	operation.sem_op = 1;
	operation.sem_flg = 0;
	
      	if(0 != semop(semaphore_id, &operation, 1))
	{
		printf("Semaphore Unlock fail\n");
	}
}

void Semaphore_Lock()
{
	struct sembuf operation;
	
	operation.sem_num = 0;
	operation.sem_op = -1;
	operation.sem_flg = 0;
	
	if(0 != semop(semaphore_id, &operation, 1))
	{
		printf("Semaphore Lock fail\n");
	}
}


