#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <unistd.h>
#include "common.h"

// 공유 메모리 접근
int ConnectSharedMemory();

// 프로세스 상태 출력
void PrintProcessStatus();

// 프로세스 종료
void StopProcess(char *name);

// 현재 프로세스를 공유메모리부터 분리
void DetachSharedMemory();

// 시그널 등록
void InitializeSignal();

// 종료 시그널 핸들러(SIGINT,SIGTERM)
void SIGINT_Handler();

void *pShemBuffer = NULL;
int shared_memory_id = 0;
int isServerExist = FALSE;

int main(int argc, char *argv[])
{
	if(argc < 2)
	{
		exit(1);
	}

	isServerExist = ConnectSharedMemory();

	if(0 == strcasecmp("status", argv[1]))
	{
		if(!isServerExist)
		{
			printf("There are no running Processes.\n");
		}
		else
		{
			PrintProcessStatus();
		}
	}
	else if(0 == strcasecmp("stop", argv[1]))
	{
		if(isServerExist)
		{
			StopProcess(argv[2]);				
		}
		else
		{
			printf("%s doesn't exist\n", argv[2]);
		}
	}	
	else if(0 == strcasecmp("start", argv[1]))
	{
		pid_t pid;

		pid = fork();

		if(pid < 0)
		{
			fprintf(stderr, "fork error\n");
			exit(1);
		}
		else if(0 == pid)
		{
			system("./server");	
		}
	}	

	if(isServerExist)
	{
		DetachSharedMemory();
	}

	return 0;
}

int ConnectSharedMemory()	
{	
	int connectSuccess = TRUE;
	
	// 공유메모리가 있는지 조사
	shared_memory_id = shmget(SHARED_MEMORY_KEY, SHARED_MEMORY_SIZE, 0644 | IPC_EXCL | IPC_CREAT);
	
	if(-1 == shared_memory_id)
	{
		// 기존 공유 메모리 접속
		shared_memory_id = shmget(SHARED_MEMORY_KEY, SHARED_MEMORY_SIZE, 0644 | IPC_CREAT);

		if(-1 == shared_memory_id)
		{
			connectSuccess = FALSE;	

			printf("Shared Memory Connect Fail\n");
	
			DetachSharedMemory();			

			exit(1);		
		}
		else
		{
			pShemBuffer = shmat(shared_memory_id, (void*)0, 0);

			if((void*)(-1) == pShemBuffer)
			{
				printf("shmat error\n");
			}
		}
	}
	else
	{
		// 기존 공유 메모리가 없으면 새로 생성된 공유메모리 해제
		if(-1 == shmctl(shared_memory_id, IPC_RMID, 0))
		{
			fprintf(stderr, "ConnectSharedMemory shmctl error\n");
		}	

		connectSuccess = FALSE;	
	}

	return connectSuccess;
}

// 현재 프로세스 상태 출력
void PrintProcessStatus()
{
	SharedData *pSharedData = (SharedData*)pShemBuffer;
	int i = 0;

	if(NULL == pSharedData)
	{
		printf("pSharedData is NULL\n");
		return;
	}

	printf("chatting server process pid = \"%d\" is running.\n", pSharedData->Server.pid);

	for(i = 0; i < pSharedData->numOfProcess; i++)
	{
		printf("chatting client process %d pid = \"%d\" is running.\n", i + 1, pSharedData->pData[i].pid);	
	}
}

// 프로세스 종료
void StopProcess(char *name)
{
	SharedData *pSharedData = (SharedData*)pShemBuffer;
	int i = 0;

	if(NULL == pSharedData)
	{
		printf("pSharedData is NULL\n");	
		return;
	}

	if(0 == strcasecmp(name, "server"))
	{
		if(0 == kill(pSharedData->Server.pid, SIGINT))
		{
			printf("server process stop signal sucess\n");
		}
		else
		{
			printf("server process stop signal fail\n");
		}
	}
	else
	{
		int found = FALSE;		

		for(i = 0; i < pSharedData->numOfProcess; i++)
		{	
			if(0 == strcasecmp(pSharedData->pData[i].name, name))
			{
				found = TRUE;
				
				if(0 == kill(pSharedData->pData[i].pid, SIGINT))
				{
					printf("client process pid [%d] stop signal success\n", pSharedData->pData[i].pid);
				}
				else
				{
					printf("client process pid [%d] stop signal fail\n", pSharedData->pData[i].pid);
				}

				break;
			}
		}

		if(!found)
		{
			printf("%s doesn't exist\n", name);
		}
	}
}

void InitializeSignal()
{
	if(SIG_ERR == signal(SIGTERM, SIGINT_Handler))
	{
		printf("cannot handle SIGTERM\n");
		exit(EXIT_FAILURE);
	}

	if(SIG_ERR == signal(SIGINT, SIGINT_Handler))
	{
		printf("cannot handle SIGINT\n");
		exit(EXIT_FAILURE);
	}
}

void DetachSharedMemory()
{
	if(-1 == shmdt(pShemBuffer))
	{
		printf("shmdt error\n");
		exit(1);
	}
}

void SIGINT_Handler()
{
	if(isServerExist)
	{
		DetachSharedMemory();
	}
	
	exit(1);
}

